# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import enum
import re
import io
import logging
import ipaddress
import numbers
from typing import (
    Match, Any, Callable, Union, Type, Tuple,
    Generic, TypeVar, Optional)
# Only used in type comments, so flake8 complains
from typing import Dict, List   # noqa: F401


_T = TypeVar('_T')
_T_contra = TypeVar('_T_contra', contravariant=True)
_E = TypeVar('_E', bound=enum.Enum)
_IPAddress = Union[ipaddress.IPv4Address, ipaddress.IPv6Address]


class Address:
    """A katcp address.

    Parameters
    ----------
    host
        Host address
    port
        Port number
    """
    __slots__ = ['_host', '_port']
    _IPV4_RE = re.compile(r'^(?P<host>[^:]+)(:(?P<port>\d+))?$')
    _IPV6_RE = re.compile(r'^\[(?P<host>[^]]+)\](:(?P<port>\d+))?$')

    def __init__(self, host: _IPAddress, port: int = None) -> None:
        self._host = host
        self._port = port

    @property
    def host(self) -> _IPAddress:
        """Host address"""
        return self._host

    @property
    def port(self) -> Optional[int]:
        """Port number"""
        return self._port

    def __str__(self) -> str:
        if isinstance(self._host, ipaddress.IPv4Address):
            prefix = str(self._host)
        else:
            prefix = '[' + str(self._host) + ']'
        if self._port is not None:
            return '{}:{}'.format(prefix, self._port)
        else:
            return prefix

    def __bytes__(self) -> bytes:
        """Encode the address for katcp protocol"""
        return str(self).encode('utf-8')

    def __repr__(self) -> str:
        if self._port is None:
            return 'Address({!r})'.format(self._host)
        else:
            return 'Address({!r}, {!r})'.format(self._host, self._port)

    @classmethod
    def parse(cls, raw: bytes) -> 'Address':
        """Construct an :class:`Address` from a katcp message argument

        Parameters
        ----------
        raw
            Unescaped value in katcp message argument

        Raises
        ------
        ValueError
            If `raw` does not represent a valid address
        """
        text = raw.decode('utf-8')
        match = cls._IPV6_RE.match(text)
        if match:
            host = ipaddress.IPv6Address(match.group('host'))  # type: _IPAddress
        else:
            match = cls._IPV4_RE.match(text)
            if match:
                host = ipaddress.IPv4Address(match.group('host'))
            else:
                raise ValueError("could not parse '{}' as an address".format(text))
        port = match.group('port')
        if port is not None:
            return cls(host, int(port))
        else:
            return cls(host)

    def __eq__(self, other):
        if not isinstance(other, Address):
            return NotImplemented
        return (self._host, self._port) == (other.host, other.port)

    def __ne__(self, other):
        return not self == other

    def __hash__(self) -> int:
        return hash((self._host, self._port))


class Timestamp(float):
    """A katcp timestamp.

    This is just a thin wrapper around :class:`float` to allow the type to be
    distinguished. It represents time in seconds as a UNIX timestamp.
    """
    pass


class Now(enum.Enum):
    """Singleton for representing a timestamp specified as ``now`` in the protocol."""
    NOW = 0


TimestampOrNow = Union[Timestamp, Now]     # type: Type[Any]


class LogLevel(enum.IntEnum):
    """katcp log level, with values matching Python log levels"""
    ALL = logging.NOTSET
    TRACE = 0
    DEBUG = logging.DEBUG
    INFO = logging.INFO
    WARN = logging.WARNING
    ERROR = logging.ERROR
    FATAL = logging.CRITICAL
    OFF = logging.CRITICAL + 10     # Higher than any level in logging module

    @classmethod
    def from_python(cls, level: int) -> 'LogLevel':
        """Map Python log level to katcp log level"""
        try:
            # Common case: value matches exactly
            return cls(level)
        except ValueError:
            # General case: round down to the next level
            ans = cls.ALL
            for member in cls.__members__.values():
                if member > ans and member <= level and member != cls.OFF:
                    ans = member
            return member


class TypeInfo(Generic[_T_contra]):
    """Type database entry. Refer to :func:`register_type` for details."""
    def __init__(self, type_: Type[_T_contra], name: str,
                 encode: Callable[[_T_contra], bytes],
                 decode: Callable[[Type[_T_contra], bytes], _T_contra],
                 default: Callable[[Type[_T_contra]], _T_contra]) -> None:
        self.type_ = type_
        self.name = name
        self.encode = encode
        self.decode = decode
        self.default = default


_types = []     # type: List[TypeInfo]


def register_type(type_: Type[_T], name: str,
                  encode: Callable[[_T], bytes],
                  decode: Callable[[Type[_T], bytes], _T],
                  default: Callable[[Type[_T]], _T] = None) -> None:
    """Register a type for encoding and decoding in messages.

    The registration is also used for subclasses of `type_` if no more
    specific registration has been made. This is particularly used for the
    registration for :class:`enum.Enum`, which is used for all enum types.

    Parameters
    ----------
    type_
        Python class.
    encode
        Function to encode values of this type to bytes
    decode
        Function to decode values of this type from bytes. It is given the
        actual derived class as the first argument.
    default
        Function to generate a default value of this type (used by the sensor
        framework). It is given the actual derived class as the first argument.
    """
    if default is None:
        default = _default_generic
    for info in _types:
        if info.type_ == type_:
            raise ValueError('{} is already registered'.format(type_))
    _types.append(TypeInfo(type_, name, encode, decode, default))


def get_type(type_: Type[_T]) -> TypeInfo[_T]:
    """Retrieve the type information previously registered with :func:`register_type`.

    It returns the last type info registered that is a superclass of `type_` (according
    to ``issubclass``.

    Raises
    ------
    TypeError
        if none of the registrations match `type_`
    """
    for info in reversed(_types):
        if issubclass(type_, info.type_):
            return info
    raise TypeError('{} is not registered'.format(type_))


def _decode_bool(cls: type, raw: bytes) -> bool:
    if raw == b'1':
        return cls(True)
    elif raw == b'0':
        return cls(False)
    else:
        raise ValueError('boolean must be 0 or 1, not {!r}'.format(raw))


def _encode_enum(value: enum.Enum) -> bytes:
    if hasattr(value, 'katcp_value'):
        return getattr(value, 'katcp_value')
    else:
        return value.name.encode('ascii').lower().replace(b'_', b'-')


def _decode_enum(cls: Type[_E], raw: bytes) -> _E:
    if hasattr(next(iter(cls)), 'katcp_value'):
        for member in cls:
            if getattr(member, 'katcp_value') == raw:
                return member
    else:
        name = raw.upper().replace(b'-', b'_').decode('ascii')
        try:
            value = cls[name]
            if raw == _encode_enum(value):
                return cls[name]
        except KeyError:
            pass
    raise ValueError('{!r} is not a valid value for {}'.format(raw, cls.__name__))


def _default_generic(cls: Type[_T]) -> _T:
    return cls()


def _default_enum(cls: Type[_E]) -> _E:
    return next(iter(cls))


# mypy doesn't allow an abstract class to be passed to Type[], hence the
# suppressions.
register_type(numbers.Real, 'float',                             # type: ignore
              lambda value: repr(float(value)).encode('ascii'),
              lambda cls, raw: cls(float(raw.decode('ascii'))))  # type: ignore
register_type(numbers.Integral, 'integer',                       # type: ignore
              lambda value: str(int(value)).encode('ascii'),
              lambda cls, raw: cls(int(raw.decode('ascii'))))    # type: ignore
register_type(bool, 'boolean',
              lambda value: b'1' if value else b'0', _decode_bool)
register_type(bytes, 'string',
              lambda value: value,
              lambda cls, raw: cls(raw))
register_type(str, 'string',
              lambda value: value.encode('utf-8'),
              lambda cls, raw: cls(raw, encoding='utf-8'))
register_type(Address, 'address',
              lambda value: bytes(value),
              lambda cls, raw: cls.parse(raw),
              lambda cls: cls(ipaddress.IPv4Address('0.0.0.0')))
register_type(Timestamp, 'timestamp',
              lambda value: repr(value).encode('ascii'),
              lambda cls, raw: cls(raw.decode('ascii')))
register_type(enum.Enum, 'discrete', _encode_enum, _decode_enum, _default_enum)


def encode(value: Any) -> bytes:
    """Encode a value to raw bytes for katcp.

    Parameters
    ----------
    value
        Value to encode

    Raises
    ------
    TypeError
        if the type of `value` has not been registered

    See also
    --------
    :func:`register_type`
    """
    return get_type(type(value)).encode(value)


def _union_args(cls: Any) -> Optional[Tuple[Type]]:
    """Convert ``Union[T1, T2]`` to (T1, T2).

    Returns ``None`` if `cls` is not a specific :class:`typing.Union` type.
    """
    if not isinstance(cls, type(Union[int, float])):
        return None
    try:
        # Python 3.5.3 onwards
        args = cls.__args__
    except AttributeError:
        args = cls.__union_params__
    return args


def decode(cls: Any, value: bytes) -> Any:
    """Decode value in katcp message to a type.

    If a union type is provided, the value must decode successfully (i.e.,
    without raising :exc:`ValueError`) for exactly one of the types in the
    union, otherwise a :exc:`ValueError` is raised.

    Parameters
    ----------
    cls
        The target type, or a :class:`typing.Union` of types.
    value
        Raw (but unescaped) value in katcp message

    Raises
    ------
    ValueError
        if `value` does not have a valid value for `cls`
    TypeError
        if `cls` is not a registered type or union of registered
        types.

    See also
    --------
    :func:`register_type`
    """
    union_args = _union_args(cls)
    if union_args is not None:
        values = []     # type: List[Any]
        for type_ in union_args:
            try:
                values.append(decode(type_, value))
            except ValueError:
                pass
        if len(values) == 1:
            return values[0]
        elif not values:
            raise ValueError('None of the types in {} could decode {}'.format(
                cls, value))
        else:
            raise ValueError('{} is ambiguous for {}'.format(value, cls))
    else:
        return get_type(cls).decode(cls, value)


class KatcpSyntaxError(ValueError):
    """Raised by parsers when encountering a syntax error."""
    def __init__(self, message: str, raw: bytes = None) -> None:
        super().__init__(message)
        self.raw = raw


class Message:
    __slots__ = ['mtype', 'name', 'arguments', 'mid']

    class Type(enum.Enum):
        """Message type"""
        REQUEST = 1
        REPLY = 2
        INFORM = 3

    _TYPE_SYMBOLS = {
        Type.REQUEST: b'?',
        Type.REPLY: b'!',
        Type.INFORM: b'#',
    }
    _REVERSE_TYPE_SYMBOLS = {
        value: key for (key, value) in _TYPE_SYMBOLS.items()}

    _NAME_RE = re.compile('^[A-Za-z][A-Za-z0-9-]*$', re.ASCII)
    _WHITESPACE_RE = re.compile(br'[ \t]+')
    _HEADER_RE = re.compile(
        br'^[!#?]([A-Za-z][A-Za-z0-9-]*)(?:\[([1-9][0-9]*)\])?$')
    #: Characters that must be escaped in an argument
    _ESCAPE_RE = re.compile(br'[\\ \0\n\r\x1b\t]')
    _UN_ESCAPE_RE = re.compile(br'\\(.)')
    #: Characters not allowed to appear in an argument
    # (space, tab are omitted because they are split on already)
    _SPECIAL_RE = re.compile(br'[\0\r\n\x1b]')

    _ESCAPE_LOOKUP = {
        b'\\': b'\\',
        b'_': b' ',
        b'0': b'\0',
        b'n': b'\n',
        b'r': b'\r',
        b'e': b'\x1b',
        b't': b'\t',
        b'@': b''
    }
    _REVERSE_ESCAPE_LOOKUP = {
        value: key for (key, value) in _ESCAPE_LOOKUP.items()}

    OK = b'ok'
    FAIL = b'fail'
    INVALID = b'invalid'

    def __init__(self, mtype: Type, name: str, *arguments: Any,
                 mid: int = None) -> None:
        self.mtype = mtype
        if not self._NAME_RE.match(name):
            raise ValueError('name {} is invalid'.format(name))
        self.name = name
        self.arguments = [encode(arg) for arg in arguments]
        if mid is not None:
            if not 1 <= mid <= 2**31 - 1:
                raise ValueError('message ID {} is outside of range 1 to 2**31-1'.format(mid))
        self.mid = mid

    @classmethod
    def request(cls, name: str, *arguments: Any, mid: int = None) -> 'Message':
        return cls(cls.Type.REQUEST, name, *arguments, mid=mid)

    @classmethod
    def reply(cls, name: str, *arguments: Any, mid: int = None) -> 'Message':
        return cls(cls.Type.REPLY, name, *arguments, mid=mid)

    @classmethod
    def inform(cls, name: str, *arguments: Any, mid: int = None) -> 'Message':
        return cls(cls.Type.INFORM, name, *arguments, mid=mid)

    @classmethod
    def reply_to_request(cls, msg: 'Message', *arguments: Any) -> 'Message':
        return cls(cls.Type.REPLY, msg.name, *arguments, mid=msg.mid)

    @classmethod
    def inform_reply(cls, msg: 'Message', *arguments: Any) -> 'Message':
        return cls(cls.Type.INFORM, msg.name, *arguments, mid=msg.mid)

    @classmethod
    def _escape_match(cls, match: Match[bytes]):
        """Given a re.Match object matching :attr:`_ESCAPE_RE`, return the escape code for it."""
        return b'\\' + cls._REVERSE_ESCAPE_LOOKUP[match.group()]

    @classmethod
    def _unescape_match(cls, match: Match[bytes]):
        char = match.group(1)
        try:
            return cls._ESCAPE_LOOKUP[char]
        except KeyError:
            raise KatcpSyntaxError('invalid escape character {!r}'.format(char))

    @classmethod
    def escape_argument(cls, arg: bytes) -> bytes:
        """Escape special bytes in an argument"""
        if arg == b'':
            return br'\@'
        else:
            return cls._ESCAPE_RE.sub(cls._escape_match, arg)

    @classmethod
    def unescape_argument(cls, arg: bytes) -> bytes:
        """Reverse of :func:`escape_argument`"""
        if arg.endswith(b'\\'):
            raise KatcpSyntaxError('argument ends with backslash')
        match = cls._SPECIAL_RE.search(arg)
        if match:
            raise KatcpSyntaxError('unescaped special {!r}'.format(match.group()))
        return cls._UN_ESCAPE_RE.sub(cls._unescape_match, arg)

    @classmethod
    def parse(cls, raw) -> 'Message':
        """Create a :class:`Message` from encoded representation.

        Parameters
        ----------
        raw
            Bytes from the wire, including the trailing newline

        Raises
        ------
        KatcpSyntaxError
            If `raw` is not validly encoded.
        """
        try:
            if not raw or raw[:1] not in b'?#!':
                raise KatcpSyntaxError('message does not start with message type')
            if raw[-1:] not in (b'\r', b'\n'):
                raise KatcpSyntaxError('message does not end with newline')
            parts = cls._WHITESPACE_RE.split(raw[:-1])
            match = cls._HEADER_RE.match(parts[0])
            if not match:
                raise KatcpSyntaxError('could not parse name and message ID')
            name = match.group(1).decode('ascii')
            mid_raw = match.group(2)
            if mid_raw is not None:
                mid = int(mid_raw)
            else:
                mid = None
            mtype = cls._REVERSE_TYPE_SYMBOLS[raw[:1]]
            # Create the message first without arguments, to avoid the argument
            # encoding and let us store raw bytes.
            msg = cls(mtype, name, mid=mid)
            # Trailing whitespace causes split to add an empty argument
            if parts[-1] == b'':
                del parts[-1]
            msg.arguments = [cls.unescape_argument(arg) for arg in parts[1:]]
            return msg
        except KatcpSyntaxError as error:
            error.raw = raw
            raise error
        except ValueError as error:
            raise KatcpSyntaxError(str(error), raw) from error

    def __bytes__(self) -> bytes:
        """Return Message as serialised for transmission"""

        output = io.BytesIO()
        output.write(self._TYPE_SYMBOLS[self.mtype])
        output.write(self.name.encode('ascii'))
        if self.mid is not None:
            output.write(b'[' + str(self.mid).encode('ascii') + b']')
        for arg in self.arguments:
            output.write(b' ')
            output.write(self.escape_argument(arg))
        output.write(b'\n')
        return output.getvalue()

    def __repr__(self) -> str:
        return ('Message(Message.Type.{self.mtype.name}, {self.name!r}').format(self=self) \
                + ''.join(', {!r}'.format(arg) for arg in self.arguments) \
                + ', mid={!r})'.format(self.mid)

    def __eq__(self, other):
        if not isinstance(other, Message):
            return NotImplemented
        for name in self.__slots__:
            if getattr(self, name) != getattr(other, name):
                return False
        return True

    def __ne__(self, other):
        return not self == other

    # Mutable, so not safely hashable.
    # The type: ignore should be removed after
    # https://github.com/python/mypy/issues/4266
    # is fixed.
    __hash__ = None      # type: ignore

    def reply_ok(self) -> bool:
        """Return True if this is a reply and its first argument is 'ok'."""
        return (self.mtype == self.Type.REPLY and bool(self.arguments) and
                self.arguments[0] == self.OK)



#!/usr/bin/env python3

# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import logging
import enum
import signal
from typing import Tuple

import aiokatcp


class Foo(enum.Enum):
    ABC_DEF = 1
    GHI_K = 2


class Server(aiokatcp.DeviceServer):
    VERSION = 'testapi-1.0'
    BUILD_STATE = 'testapi-1.0.1'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        sensor = aiokatcp.Sensor(int, 'counter-queries', 'number of ?counter queries',
                                 default=0,
                                 initial_status=aiokatcp.Sensor.Status.NOMINAL)
        self.sensors.add(sensor)
        sensor = aiokatcp.Sensor(Foo, 'foo', 'nonsense')
        self.sensors.add(sensor)

    async def request_echo(self, ctx, *args: str) -> Tuple:
        """Return the arguments to the caller"""
        return tuple(args)

    async def request_sleep(self, ctx, time: float) -> None:
        """Sleep for some amount of time"""
        await asyncio.sleep(time, loop=self.loop)

    async def request_fail(self, ctx, arg: str) -> None:
        """Request that always returns a failure reply"""
        raise aiokatcp.FailReply(arg + ' is no good')

    async def request_crash(self, ctx) -> None:
        """Request that always raises an exception"""
        raise RuntimeError("help I've fallen over and can't get up")

    async def request_counter(self, ctx) -> None:
        """Increment counter-queries"""
        self.sensors['counter-queries'].value += 1


async def main():
    server = Server('localhost', 4444)
    handler = Server.LogHandler(server)
    logging.getLogger().addHandler(handler)
    await server.start()
    asyncio.get_event_loop().add_signal_handler(signal.SIGINT, server.halt)
    await server.join()


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    loop.run_until_complete(main())
    loop.close()



# Copyright 2017, 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import logging
import re
import warnings
import inspect
import random
import functools
import enum
import time
import contextlib
from collections import OrderedDict
from typing import Any, List, Iterable, Callable, Tuple, Sequence, Type, Set, Generator
# Only used in type comments, so flake8 complains
from typing import Dict, Optional, Union   # noqa: F401

from . import core, connection, sensor
from .connection import FailReply, InvalidReply


logger = logging.getLogger(__name__)
_InformHandler = Callable[['Client', core.Message], None]


class _PendingRequest:
    def __init__(self, name: str, mid: Optional[int], loop: asyncio.AbstractEventLoop) -> None:
        self.name = name
        self.mid = mid
        self.informs = []      # type: List[core.Message]
        self.reply = loop.create_future()


class ClientMeta(type):
    @classmethod
    def _wrap_inform(mcs, name: str, value: Callable[..., None]) -> _InformHandler:
        return connection.wrap_handler(name, value, 1)

    def __new__(mcs, name, bases, namespace, **kwds):
        namespace.setdefault('_inform_handlers', {})
        for base in bases:
            namespace['_inform_handlers'].update(getattr(base, '_inform_handlers', {}))
        result = type.__new__(mcs, name, bases, namespace)
        inform_handlers = getattr(result, '_inform_handlers')
        for key, value in namespace.items():
            if key.startswith('inform_') and inspect.isfunction(value):
                request_name = key[7:].replace('_', '-')
                inform_handlers[request_name] = mcs._wrap_inform(request_name, value)
        return result


def _make_done(future):
    if not future.done():
        future.set_result(None)


class ProtocolError(ValueError):
    """The server does not implement the required protocol version"""
    def __init__(self, msg, version):
        super().__init__(msg)
        self.version = version


class Client(metaclass=ClientMeta):
    """Client that connects to a katcp server.

    The client will automatically connect to the server, and reconnect if
    the connection is lost. If you want to wait for the initial connection
    to complete up front, the :meth:`.connect` factory may be preferable to the
    constructor.

    Parameters
    ----------
    host
        Server hostname
    port
        Server port number
    limit
        Maximum line length in a message from the server
    loop
        Event loop on which the client will run, defaulting to
        ``asyncio.get_event_loop()``.

    Attributes
    ----------
    is_connected : bool
        Whether the connection is currently established.
    last_exc : Exception
        An exception object associated with the last connection attempt. It is
        always ``None`` if :attr:`is_connected` is True.
    """
    def __init__(self, host: str, port: int, *,
                 auto_reconnect: bool = True,
                 limit: int = connection.DEFAULT_LIMIT,
                 loop: asyncio.AbstractEventLoop = None) -> None:
        if loop is None:
            loop = asyncio.get_event_loop()
        self._connection = None         # type: Optional[connection.Connection]
        self.is_connected = False
        self.host = host
        self.port = port
        self.loop = loop
        self.logger = logger     # type: Union[logging.Logger, connection.ConnectionLoggerAdapter]
        self._limit = limit
        self._pending = {}              # type: Dict[Optional[int], _PendingRequest]
        self._next_mid = 1
        self._run_task = loop.create_task(self._run())
        self._run_task.add_done_callback(self._done_callback)
        self._closing = False
        self._closed_event = asyncio.Event(loop=loop)
        self._connected_callbacks = []       # type: List[Callable[[], None]]
        self._disconnected_callbacks = []    # type: List[Callable[[], None]]
        self._failed_connect_callbacks = []  # type: List[Callable[[Exception], None]]
        self._inform_callbacks = {}      # type: Dict[str, List[Callable[[core.Message], None]]]
        self._sensor_monitor = None          # type: Optional[_SensorMonitor]
        self._mid_support = False
        # Used to serialize requests if the server does not support message IDs
        self._request_lock = asyncio.Lock(loop=loop)
        self.auto_reconnect = auto_reconnect
        if self.auto_reconnect:
            # If not auto-reconnecting, wait_connected will set the exception
            self.add_failed_connect_callback(self._warn_failed_connect)
        self.last_exc = None                 # type: Optional[Exception]

    def __del__(self) -> None:
        if hasattr(self, '_closed_event') and not self._closed_event.is_set():
            warnings.warn('unclosed Client {!r}'.format(self), ResourceWarning)
            if not self.loop.is_closed():
                self.loop.call_soon_threadsafe(self.close)

    def _set_connection(self, conn: Optional[connection.Connection]):
        self._connection = conn
        if conn is None:
            self.logger = logger
        else:
            self.logger = connection.ConnectionLoggerAdapter(
                logger, dict(address=conn.address))

    async def handle_message(self, conn: connection.Connection, msg: core.Message) -> None:
        """Called by :class:`~.Connection` for each incoming message."""
        if msg.mtype == core.Message.Type.REQUEST:
            self.logger.info('Received unexpected request %s from server', msg.name)
            return
        if msg.mid is not None or (not self._mid_support
                                   and None in self._pending
                                   and self._pending[None].name == msg.name):
            try:
                req = self._pending[msg.mid]
            except KeyError:
                self.logger.debug(
                    'Received %r with unknown message ID %s (possibly cancelled request)',
                    bytes(msg), msg.mid)
            else:
                if msg.mtype == core.Message.Type.REPLY:
                    if not req.reply.done():
                        req.reply.set_result(msg)
                elif msg.mtype == core.Message.Type.INFORM:
                    req.informs.append(msg)
                else:
                    self.logger.warning('Unknown message type %s', msg.mtype)  # pragma: no cover
        elif msg.mtype == core.Message.Type.INFORM:
            self.handle_inform(msg)
        else:
            self.logger.info('Received unexpected %s (%s) from server without message ID',
                             msg.mtype.name, msg.name)

    def handle_inform(self, msg: core.Message) -> None:
        self.logger.debug('Received %s', bytes(msg))
        # TODO: provide dispatch mechanism for informs
        handler = self._inform_handlers.get(               # type: ignore
            msg.name, self.__class__.unhandled_inform)
        try:
            handler(self, msg)
        except FailReply as error:
            self.logger.warning('error in inform %s: %s', msg.name, error)
        except Exception:
            self.logger.exception('unhandled exception in inform %s', msg.name, exc_info=True)
        self._run_callbacks(self._inform_callbacks.get(msg.name, {}), msg)

    def unhandled_inform(self, msg: core.Message) -> None:
        """Called if an inform is received for which no handler is registered.

        The default simply logs a debug message if there are no inform
        callbacks registered for the message. Subclasses may override this to
        provide other behaviour for unknown informs.
        """
        if msg.name not in self._inform_callbacks:
            self.logger.debug('unknown inform %s', msg.name)

    def _close_connection(self) -> None:
        if self._connection is not None:
            self._connection.close()
            self._set_connection(None)

    def _warn_failed_connect(self, exc: Exception) -> None:
        self.logger.warning('Failed to connect to %s:%s: %s', self.host, self.port, exc)

    def inform_version_connect(self, api: str, version: str, build_state: str = None) -> None:
        if api == 'katcp-protocol':
            match = re.match(r'^(\d+)\.(\d+)(?:-(.+))?$', version)
            error = None
            if not match:
                error = 'Unparsable katcp-protocol {!r}'.format(version)
            else:
                major = int(match.group(1))
                minor = int(match.group(2))
                self.logger.debug('Protocol version %d.%d', major, minor)
                flags = match.group(3)
                if major != 5:
                    error = 'Unknown protocol version {}.{}'.format(major, minor)
            if error is None:
                self._mid_support = (flags is not None and 'I' in flags)
                # Safety in case a race condition causes the connection to
                # die before this function was called.
                if self._connection is not None:
                    self._on_connected()
            else:
                self._close_connection()
                self._on_failed_connect(ProtocolError(error, version))
        # TODO: add a inform_version handler

    def inform_disconnect(self, reason: str) -> None:
        self.logger.info('Server disconnected: %s', reason)
        self._close_connection()

    def add_connected_callback(self, callback: Callable[[], None]) -> None:
        """Register a handler that is called when a connection is established.

        The handler is called without arguments. Use a lambda or
        :func:`functools.partial` if you need arguments. Handlers are called in
        the order they are registered.
        """
        self._connected_callbacks.append(callback)

    def remove_connected_callback(self, callback: Callable[[], None]) -> None:
        """Remove a callback registered with :meth:`add_connected_callback`."""
        self._connected_callbacks.remove(callback)

    def add_disconnected_callback(self, callback: Callable[[], None]) -> None:
        """Register a handler that is called when a connection is lost.

        The handler is called without arguments. Use a lambda or
        :func:`functools.partial` if you need arguments. Handlers are called in
        the reverse of order of registration.
        """
        self._disconnected_callbacks.append(callback)

    def remove_disconnected_callback(self, callback: Callable[[], None]) -> None:
        """Remove a callback registered with :meth:`add_disconnected_callback`."""
        self._disconnected_callbacks.remove(callback)

    def add_failed_connect_callback(self, callback: Callable[[Exception], None]) -> None:
        """Register a handler that is called when a connection attempt fails.

        The handler is passed an exception object. Handlers are called in the
        order of registration.
        """
        self._failed_connect_callbacks.append(callback)

    def remove_failed_connect_callback(self, callback: Callable[[Exception], None]) -> None:
        """Remove a callback registered with :meth:`add_failed_connect_callback`."""
        self._failed_connect_callbacks.remove(callback)

    def add_inform_callback(self, name: str, callback: Callable[..., None]) -> None:
        """Add a callback called on every asynchronous inform.

        The message arguments are unpacked according to the type annotations
        on the arguments of the callback. Callbacks are called in the order
        registered, after any handlers defined by methods in the class.
        """
        wrapper = connection.wrap_handler(name, callback, 0)
        self._inform_callbacks.setdefault(name, []).append(wrapper)

    def remove_inform_callback(self, name: str, callback: Callable[..., None]) -> None:
        """Remove a callback registered with :meth:`add_inform_callback`."""
        cbs = self._inform_callbacks.get(name, [])
        for i in range(len(cbs)):
            if cbs[i]._aiokatcp_orig_handler == callback:     # type: ignore
                del cbs[i]
                if not cbs:
                    del self._inform_callbacks[name]
                break

    # callbacks should be marked as Iterable[Callable[..., None]], but in
    # Python 3.5.2 that gives an error in the typing module.
    def _run_callbacks(self, callbacks: Iterable, *args) -> None:
        # Wrap in list() so that the callbacks can safely mutate the original
        for callback in list(callbacks):
            try:
                callback(*args)
            except Exception:
                self.logger.exception('Exception raised from callback')

    def _on_connected(self) -> None:
        if not self.is_connected:
            self.is_connected = True
            self.last_exc = None
            self._run_callbacks(self._connected_callbacks)

    def _on_disconnected(self) -> None:
        if self.is_connected:
            self.is_connected = False
            self.last_exc = ConnectionResetError('Connection to server lost')
            for req in self._pending.values():
                if not req.reply.done():
                    req.reply.set_exception(self.last_exc)
            self._pending.clear()
            self._run_callbacks(reversed(self._disconnected_callbacks))

    def _on_failed_connect(self, exc: Exception) -> None:
        self.last_exc = exc
        self._run_callbacks(self._failed_connect_callbacks, exc)

    async def _run_once(self) -> bool:
        """Make a single attempt to connect and run the connection if successful."""
        # Open the connection. Based on asyncio.open_connection.
        reader = asyncio.StreamReader(limit=self._limit, loop=self.loop)
        protocol = connection.ConvertCRProtocol(reader, loop=self.loop)
        try:
            transport, _ = await self.loop.create_connection(
                lambda: protocol, self.host, self.port)
        except OSError as error:
            self._on_failed_connect(error)
            return False
        writer = asyncio.StreamWriter(transport, protocol, reader, self.loop)
        conn = connection.Connection(self, reader, writer, False)
        self._set_connection(conn)
        # Process replies until connection closes. _on_connected is
        # called by the version-info inform handler.
        await conn.wait_closed()
        ret = self.is_connected
        if self.is_connected:
            self._on_disconnected()
        return ret

    async def _run(self) -> None:
        if self.auto_reconnect:
            backoff = 0.5
            while True:
                success = await self._run_once()
                if success:
                    backoff = 1.0
                else:
                    # Exponential backoff if connections are failing
                    backoff = min(backoff * 2.0, 60.0)
                # Pick a random value in [0.5 * backoff, backoff]
                wait = (random.random() * 1.0) * 0.5 * backoff
                await asyncio.sleep(wait, loop=self.loop)
        else:
            await self._run_once()

    def _done_callback(self, future: asyncio.Future) -> None:
        self._close_connection()
        if self.is_connected:
            self._on_disconnected()
        else:
            self._on_failed_connect(ConnectionAbortedError('close() called'))
        self._closed_event.set()

    def close(self) -> None:
        """Start closing the connection.

        Closing completes asynchronously. Use :meth:`wait_closed` to wait
        for it to be fully complete.
        """
        # TODO: also needs to abort any pending requests
        if not self._closing:
            self._run_task.cancel()
            self._close_connection()    # Ensures the transport gets closed now
            self._closing = True
            if self._sensor_monitor is not None:
                self._sensor_monitor.close(True)
                self._sensor_monitor = None

    async def wait_closed(self) -> None:
        """Wait for the process started by :meth:`close` to complete."""
        await self._closed_event.wait()

    # Make client a context manager that self-closes
    async def __aenter__(self) -> 'Client':
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        self.close()
        await self.wait_closed()

    def _set_last_exc(self, future: asyncio.Future, exc: Exception) -> None:
        if not future.done():
            future.set_exception(exc)

    async def wait_connected(self) -> None:
        """Wait until a connection is established.

        If construct with ``auto_reconnect=False``, then this will raise an
        exception if the single connection attempt failed. Otherwise, it will
        block indefinitely until a connection is successful.

        .. note::

            On return, it is possible that :attr:`is_connected` is false,
            because the connection may fail immediately after waking up the
            waiter.
        """
        if not self.is_connected:
            if not self.auto_reconnect and self.last_exc is not None:
                # This includes the case of having had a successful connection
                # that disconnected.
                raise self.last_exc
            future = self.loop.create_future()
            callback = functools.partial(_make_done, future)
            if self.auto_reconnect:
                failed_callback = None   # type: Optional[Callable[[Exception], None]]
            else:
                failed_callback = functools.partial(self._set_last_exc, future)
                self.add_failed_connect_callback(failed_callback)
            self.add_connected_callback(callback)
            try:
                await future
            finally:
                self.remove_connected_callback(callback)
                if failed_callback:
                    self.remove_failed_connect_callback(failed_callback)

    async def wait_disconnected(self) -> None:
        """Wait until there is no connection"""
        if self.is_connected:
            future = self.loop.create_future()
            callback = functools.partial(_make_done, future)
            self.add_disconnected_callback(callback)
            try:
                await future
            finally:
                self.remove_disconnected_callback(callback)

    @classmethod
    async def connect(cls, host: str, port: int, *,
                      auto_reconnect: bool = True,
                      limit: int = connection.DEFAULT_LIMIT,
                      loop: asyncio.AbstractEventLoop = None) -> 'Client':
        """Factory function that creates a client and waits until it is connected.

        Refer to the constructor documentation for details of the parameters.
        """
        client = cls(host, port, auto_reconnect=auto_reconnect, limit=limit, loop=loop)
        try:
            await client.wait_connected()
            return client
        except Exception as error:
            client.close()
            raise error

    async def request_raw(self, name: str, *args: Any) -> Tuple[core.Message, List[core.Message]]:
        """Make a request to the server and await the reply, without decoding it.

        Parameters
        ----------
        name
            Message name
        args
            Message arguments, which will be encoded by :class:`~.core.Message`.

        Returns
        -------
        reply
            Reply message
        informs
            List of synchronous informs received

        Raises
        ------
        BrokenPipeError
            if not connected at the time the request was made
        ConnectionError
            if the connection was lost before the reply was received
        """
        if not self.is_connected:
            raise BrokenPipeError('Not connected')
        if self._mid_support:
            mid = self._next_mid
            self._next_mid += 1
            return await self._request_raw_impl(mid, name, *args)
        else:
            async with self._request_lock:
                return await self._request_raw_impl(None, name, *args)

    async def _request_raw_impl(self, mid: Optional[int], name: str, *args: Any) -> \
            Tuple[core.Message, List[core.Message]]:
        req = _PendingRequest(name, mid, self.loop)
        self._pending[mid] = req
        try:
            msg = core.Message(core.Message.Type.REQUEST, name, *args, mid=mid)
            assert self._connection is not None
            self._connection.write_message(msg)
            reply_msg = await req.reply
            return reply_msg, req.informs
        finally:
            self._pending.pop(mid, None)

    async def request(self, name: str, *args: Any) -> Tuple[List[bytes], List[core.Message]]:
        """Make a request to the server and await the reply.

        It expects the first argument of the reply to be ``ok``, ``fail`` or
        ``invalid``, and raises exceptions in the latter two cases. If this is
        undesirable, use :meth:`request_raw` instead.

        Parameters
        ----------
        name
            Message name
        args
            Message arguments, which will be encoded by :class:`~.core.Message`.

        Returns
        -------
        reply
            Reply message arguments, excluding the ``ok``
        informs
            List of synchronous informs received

        Raises
        ------
        FailReply
            if the server replied with ``fail``
        InvalidReply
            if the server replied anything except ``ok`` or ``fail``
        BrokenPipeError
            if not connected at the time the request was made
        ConnectionError
            if the connection was lost before the reply was received
        """
        reply_msg, informs = await self.request_raw(name, *args)
        type_ = core.Message.INVALID if not reply_msg.arguments else reply_msg.arguments[0]
        error = b'' if len(reply_msg.arguments) <= 1 else reply_msg.arguments[1]
        if type_ == core.Message.OK:
            return reply_msg.arguments[1:], informs
        elif type_ == core.Message.FAIL:
            raise FailReply(error.decode('utf-8', errors='replace'))
        else:
            raise InvalidReply(error.decode('utf-8', errors='replace'))

    def add_sensor_watcher(self, watcher: 'AbstractSensorWatcher') -> None:
        if self._sensor_monitor is None:
            self._sensor_monitor = _SensorMonitor(self)
        self._sensor_monitor.add_watcher(watcher)

    def remove_sensor_watcher(self, watcher: 'AbstractSensorWatcher') -> None:
        if self._sensor_monitor is not None:
            self._sensor_monitor.remove_watcher(watcher)
            if not self._sensor_monitor:   # i.e. no more watchers
                self._sensor_monitor.close(False)
                self._sensor_monitor = None


class SyncState(enum.Enum):
    """State of synchronisation of an :class:`AbstractSensorWatcher`"""
    #: Not currently connected to the server
    DISCONNECTED = 1
    #: Connected to the server, but still subscribing to sensors
    SYNCING = 2
    #: Connected to the server and sensor list is up to date
    SYNCED = 3
    #: Client object has been closed (:meth:`Client.close`)
    CLOSED = 4


class AbstractSensorWatcher:
    """Base class for receiving notifications about sensor changes.

    This class is intended to be subclassed to implement any of the
    notification callbacks.
    """
    def sensor_added(self, name: str, description: str, units: str, type_name: str,
                     *args: bytes) -> None:
        """A sensor was added on the remote server.

        This is also called if a sensor changed its properties. In that case
        there is *no* call to :meth:`sensor_removed`.
        """
        pass         # pragma: nocover

    def sensor_removed(self, name: str) -> None:
        """A sensor disappeared from the remote server."""
        pass         # pragma: nocover

    def sensor_updated(self, name: str, value: bytes,
                       status: sensor.Sensor.Status, timestamp: float) -> None:
        """The value of a sensor changed on the remote server."""
        pass         # pragma: nocover

    def batch_start(self) -> None:
        """Called at the start of a batch of back-to-back updates.

        Calls to :meth:`sensor_added`, :meth:`sensor_removed` and :meth:`sensor_updated`
        will always be bracketed by :meth:`batch_start` and :meth:`batch_stop`. This
        does not apply to :meth:`state_updated`."""
        pass         # pragma: nocover

    def batch_stop(self) -> None:
        """Called at the end of a batch of back-to-back updates."""
        pass         # pragma: nocover

    def state_updated(self, state: SyncState) -> None:
        """Indicates the state of the synchronisation state machine.

        Implementations should assume the initial state is
        :const:`SyncState.DISCONNECTED`.
        """
        pass         # pragma: nocover


class DiscreteMixin:
    @property
    def katcp_value(self):
        return self.value


class SensorWatcher(AbstractSensorWatcher):
    """Sensor watcher that mirrors sensors into a :class:`SensorSet`.

    Parameters
    ----------
    client
        Client to which this watcher will be attached. It is currently only used to
        get the correct logger and event loop.
    enum_types
        Enum types to be used for discrete sensors. An enum type is used if it
        has the same legal values in the same order as the remote sensor. If
        a discrete sensor has no matching enum type, one is synthesized on the
        fly.

    Attributes
    ----------
    sensors : :class:`SensorSet`
        The mirrored sensors
    synced : :class:`asyncio.Event`
        Event that is set whenever the state is :const:`SyncState.SYNCED`
    """

    SENSOR_TYPES = {
        'integer': int,
        'float': float,
        'boolean': bool,
        'timestamp': core.Timestamp,
        'discrete': enum.Enum,    # Actual type is constructed dynamically
        'address': core.Address,
        'string': bytes      # Allows passing through arbitrary values even if not UTF-8
    }

    def __init__(self, client: Client, enum_types: Sequence[Type[enum.Enum]] = ()) -> None:
        self.synced = asyncio.Event(loop=client.loop)
        self.logger = client.logger
        self.sensors = sensor.SensorSet()
        # Synthesized enum types for discrete sensors
        self._enum_cache = {}              # type: Dict[Tuple[bytes, ...], Type[enum.Enum]]
        for enum_type in enum_types:
            key = tuple(core.encode(value) for value in enum_type.__members__.values())
            self._enum_cache[key] = enum_type

    def rewrite_name(self, name: str) -> str:
        """Convert name of incoming sensor to name to use in the sensor set.

        This defaults to the identity, but can be overridden to provide name mangling.
        """
        return name

    def make_type(self, type_name: str, parameters: Sequence[bytes]) -> type:
        """Get the sensor type for a given type name"""
        if type_name == 'discrete':
            values = tuple(parameters)
            if values in self._enum_cache:
                return self._enum_cache[values]
            else:
                # We need unique Python identifiers for each value, but simply
                # normalising names in some way doesn't guarantee that.
                # Instead, we use arbitrary numbering.
                enums = [('ENUM{}'.format(i), value) for i, value in enumerate(values)]
                # Type checking disabled due to https://github.com/python/mypy/issues/4184
                stype = enum.Enum('discrete', enums, type=DiscreteMixin)  # type: ignore
                self._enum_cache[values] = stype
                return stype
        else:
            return self.SENSOR_TYPES[type_name]

    def sensor_added(self, name: str, description: str, units: str, type_name: str,
                     *args: bytes) -> None:
        if type_name not in self.SENSOR_TYPES:
            self.logger.warning('Type %s is not recognised, skipping sensor %s',
                                type_name, name)
            return
        stype = self.make_type(type_name, args)
        s = sensor.Sensor(stype, self.rewrite_name(name),
                          description, units)   # type: sensor.Sensor
        self.sensors.add(s)

    def sensor_removed(self, name: str) -> None:
        self.sensors.pop(self.rewrite_name(name), None)

    def sensor_updated(self, name: str, value: bytes,
                       status: sensor.Sensor.Status, timestamp: float) -> None:
        try:
            sensor = self.sensors[self.rewrite_name(name)]
        except KeyError:
            self.logger.warning('Received update for unknown sensor %s', name)
            return

        try:
            decoded = core.decode(sensor.stype, value)
        except ValueError as exc:
            self.logger.warning('Sensor %s: value %r does not match type %s: %s',
                                name, value, sensor.type_name, exc)
            return

        sensor.set_value(decoded, status=status, timestamp=timestamp)

    def state_updated(self, state: SyncState) -> None:
        if state == SyncState.DISCONNECTED:
            now = time.time()
            for s in self.sensors.values():
                s.set_value(s.value, status=sensor.Sensor.Status.UNREACHABLE, timestamp=now)

        if state == SyncState.SYNCED:
            self.synced.set()
        else:
            self.synced.clear()


class _SensorMonitor:
    """Tracks the sensors on a client.

    Only a single instance is added to a given client, and it distributes
    notifications to instances of :class:`SensorWatcher`.

    Users should not interact with this class directly.
    """

    def __init__(self, client: Client) -> None:
        self.client = client
        self.logger = client.logger
        client.add_connected_callback(self._connected)
        client.add_disconnected_callback(self._disconnected)
        client.add_inform_callback('interface-changed', self._interface_changed)
        client.add_inform_callback('sensor-status', self._sensor_status)
        self._update_task = None           # type: Optional[asyncio.Task]
        # Sensors we have seen: maps name to arguments
        self._sensors = {}                 # type: Dict[str, Tuple[bytes, ...]]
        # Sensors whose sampling strategy has been set
        self._sampling_set = set()         # type: Set[str]
        self._in_batch = False
        # Really an OrderedSet, but no such type exists
        self._watchers = OrderedDict()     # type: Dict[AbstractSensorWatcher, None]

    def add_watcher(self, watcher: AbstractSensorWatcher) -> None:
        self._watchers[watcher] = None

    def remove_watcher(self, watcher: AbstractSensorWatcher) -> None:
        try:
            del self._watchers[watcher]
        except KeyError:
            pass

    def __bool__(self) -> bool:
        """True if there are any watchers"""
        return bool(self._watchers)

    @contextlib.contextmanager
    def _batch(self) -> Generator[None, None, None]:
        assert not self._in_batch, 'Re-entered _batch'
        self._in_batch = True
        self.logger.debug('Entering batch')
        try:
            for watcher in self._watchers:
                watcher.batch_start()
            yield
        finally:
            self.logger.debug('Exiting batch')
            for watcher in self._watchers:
                watcher.batch_stop()
            self._in_batch = False

    def _cancel_update(self) -> None:
        if self._update_task is not None:
            self._update_task.cancel()
            self._update_task = None

    def _update_done(self, future):
        try:
            future.result()
        except asyncio.CancelledError:
            pass
        except OSError as error:
            # Connection died before we finished. Log it, but no need for
            # a stack trace.
            self.logger.warning('Connection error in update task: %s', error)
        except Exception:
            self.logger.exception('Exception in update task')

    def _trigger_update(self) -> None:
        self.logger.debug('Sensor sync triggered')
        for watcher in self._watchers:
            watcher.state_updated(SyncState.SYNCING)
        self._cancel_update()
        self._update_task = self.client.loop.create_task(self._update())
        self._update_task.add_done_callback(self._update_done)

    async def _set_sampling(self, names: Sequence[str]) -> None:
        """Register sampling strategy with sensors in `names`"""
        coros = [self.client.request('sensor-sampling', name, 'auto') for name in names]
        results = await asyncio.gather(*coros, loop=self.client.loop, return_exceptions=True)
        for name, result in zip(names, results):
            if isinstance(result, Exception):
                try:
                    raise result
                except (FailReply, InvalidReply) as error:
                    self.logger.warning('Failed to set strategy on %s: %s', name, error)
            else:
                self._sampling_set.add(name)

    async def _update(self) -> None:
        """Refresh the sensor list and subscriptions."""
        reply, informs = await self.client.request('sensor-list')
        sampling = []       # type: List[str]
        seen = set()        # type: Set[str]
        with self._batch():
            # Enumerate all sensors and add new or changed ones
            for inform in informs:
                name, description, units, type_name = [
                    core.decode(str, inform.arguments[i]) for i in range(4)]
                seen.add(name)
                params = tuple(inform.arguments[1:])
                # Check if it already exists with the same parameters
                old = self._sensors.get(name)
                if old != params:
                    for watcher in self._watchers:
                        watcher.sensor_added(name, description, units, type_name,
                                             *inform.arguments[4:])
                    self._sensors[name] = params
                    self._sampling_set.discard(name)
                if name not in self._sampling_set:
                    sampling.append(name)
            # Remove old sensors
            for name in list(self._sensors.keys()):
                if name not in seen:
                    for watcher in self._watchers:
                        watcher.sensor_removed(name)
                    self._sampling_set.discard(name)
        await self._set_sampling(sampling)
        for watcher in self._watchers:
            watcher.state_updated(SyncState.SYNCED)

    async def _unsubscribe(self, sampling_set: Set[str]) -> None:
        for name in sampling_set:
            await self.client.request('sensor-sampling', name, 'none')

    def _connected(self) -> None:
        self._sampling_set.clear()
        self._trigger_update()

    def _disconnected(self) -> None:
        self._sampling_set.clear()
        self._cancel_update()
        for watcher in self._watchers:
            watcher.state_updated(SyncState.DISCONNECTED)

    def _interface_changed(self, *args: bytes) -> None:
        # This could eventually be smarter and consult the args
        self._trigger_update()

    def _sensor_status(self, timestamp: core.Timestamp, n: int, *args) -> None:
        if len(args) != 3 * n:
            raise FailReply('Incorrect number of arguments')
        with self._batch():
            for i in range(n):
                name = '<unknown>'
                try:
                    name = core.decode(str, args[3 * i])
                    status = core.decode(sensor.Sensor.Status, args[3 * i + 1])
                    value = args[3 * i + 2]
                    for watcher in self._watchers:
                        watcher.sensor_updated(name, value, status, timestamp)
                except Exception:
                    self.logger.warning('Failed to process #sensor-status for %s',
                                        name, exc_info=True)

    def close(self, client_closing: bool) -> None:
        self._cancel_update()
        self.client.remove_connected_callback(self._connected)
        self.client.remove_disconnected_callback(self._disconnected)
        self.client.remove_inform_callback('interface-changed', self._interface_changed)
        self.client.remove_inform_callback('sensor-status', self._sensor_status)
        # The monitor is closed if there are no more watchers or if the
        # client is closed. In the latter case, let the watchers know.
        for watcher in self._watchers:
            watcher.state_updated(SyncState.CLOSED)
        if not client_closing and self._sampling_set:
            task = self.client.loop.create_task(self._unsubscribe(set(self._sampling_set)))
            task.add_done_callback(self._update_done)
        self._sensors.clear()
        self._sampling_set.clear()



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for :mod:`aiokatcp.sensor`. Most of the testing is done indirectly
in :mod:`aiokatcp.test.test_server`.
"""

import gc
import unittest

import asynctest

from aiokatcp.sensor import Sensor, SensorSampler, SensorSet


class TestSensorState(unittest.TestCase):
    def test_valid_value(self):
        Status = Sensor.Status
        self.assertFalse(Status.UNKNOWN.valid_value())
        self.assertTrue(Status.NOMINAL.valid_value())
        self.assertTrue(Status.WARN.valid_value())
        self.assertTrue(Status.ERROR.valid_value())
        self.assertFalse(Status.FAILURE.valid_value())
        self.assertFalse(Status.UNREACHABLE.valid_value())
        self.assertFalse(Status.INACTIVE.valid_value())


class TestSensor(unittest.TestCase):
    def test_status_func(self):
        def status_func(value):
            return Sensor.Status.WARN if value & 1 else Sensor.Status.ERROR

        sensor = Sensor(int, 'sensor', status_func=status_func)
        self.assertEqual(sensor.status, Sensor.Status.UNKNOWN)
        sensor.value = 1
        self.assertEqual(sensor.status, Sensor.Status.WARN)
        sensor.value = 2
        self.assertEqual(sensor.status, Sensor.Status.ERROR)
        sensor.set_value(1, Sensor.Status.NOMINAL)
        self.assertEqual(sensor.status, Sensor.Status.NOMINAL)


class TestSensorSampling(asynctest.TestCase):
    async def test_unclosed_sampler(self):
        sensor = Sensor(int, 'sensor')
        sampler = SensorSampler.factory(
            sensor, lambda sensor, reading: None,
            self.loop, SensorSampler.Strategy.EVENT)
        with self.assertWarns(ResourceWarning):
            del sensor
            del sampler
            # Run gc twice because PyPy sometimes needs this.
            gc.collect()
            gc.collect()


class TestSensorSet(unittest.TestCase):
    def setUp(self):
        self.ss = SensorSet()
        self.sensors = [Sensor(int, 'name{}'.format(i)) for i in range(5)]
        # A different set of sensors with the same names
        self.alt_sensors = [Sensor(float, 'name{}'.format(i)) for i in range(5)]
        self.ss.add(self.sensors[0])
        self.add_callback = unittest.mock.MagicMock()
        self.remove_callback = unittest.mock.MagicMock()
        self.ss.add_add_callback(self.add_callback)
        self.ss.add_remove_callback(self.remove_callback)

    def _assert_sensors(self, ss, sensors):
        """Assert that `ss` has the same sensors as `sensors`"""
        ordered = sorted(ss.values(), key=lambda x: x.name)
        self.assertEqual(ordered, sensors)

    def test_construct(self):
        """Test that setUp put things into the right state."""
        self._assert_sensors(self.ss, [self.sensors[0]])

    def test_add(self):
        # Add a new one
        self.ss.add(self.sensors[1])
        self._assert_sensors(self.ss, [self.sensors[0], self.sensors[1]])
        self.add_callback.assert_called_with(self.sensors[1])
        # Add the same one
        self.add_callback.reset_mock()
        self.ss.add(self.sensors[0])
        self._assert_sensors(self.ss, [self.sensors[0], self.sensors[1]])
        self.add_callback.assert_not_called()
        # Replace one
        self.remove_callback.assert_not_called()
        self.ss.add(self.alt_sensors[1])
        self._assert_sensors(self.ss, [self.sensors[0], self.alt_sensors[1]])
        self.remove_callback.assert_called_once_with(self.sensors[1])
        self.add_callback.assert_called_once_with(self.alt_sensors[1])

    def test_remove(self):
        # Try to remove non-existent name
        with self.assertRaises(KeyError):
            self.ss.remove(self.sensors[4])
        # Try to remove one with the same name as an existing one
        with self.assertRaises(KeyError):
            self.ss.remove(self.alt_sensors[0])
        self._assert_sensors(self.ss, [self.sensors[0]])
        # Remove one
        self.remove_callback.assert_not_called()
        self.ss.remove(self.sensors[0])
        self._assert_sensors(self.ss, [])
        self.remove_callback.assert_called_once_with(self.sensors[0])

    def test_discard(self):
        # Try to remove non-existent name
        self.ss.discard(self.sensors[4])
        # Try to remove one with the same name as an existing one
        self.ss.discard(self.alt_sensors[0])
        self._assert_sensors(self.ss, [self.sensors[0]])
        # Remove one
        self.remove_callback.assert_not_called()
        self.ss.discard(self.sensors[0])
        self._assert_sensors(self.ss, [])
        self.remove_callback.assert_called_once_with(self.sensors[0])

    def test_clear(self):
        self.ss.add(self.sensors[1])
        self.ss.clear()
        self._assert_sensors(self.ss, [])
        self.remove_callback.assert_any_call(self.sensors[0])
        self.remove_callback.assert_any_call(self.sensors[1])

    def test_popitem(self):
        self.ss.add(self.sensors[1])
        items = []
        try:
            for _ in range(100):   # To prevent infinite loop if it's broken
                items.append(self.ss.popitem())
        except KeyError:
            pass
        items.sort(key=lambda x: x[0])
        self.assertEqual(items, [('name0', self.sensors[0]), ('name1', self.sensors[1])])
        self.remove_callback.assert_any_call(self.sensors[0])
        self.remove_callback.assert_any_call(self.sensors[1])

    def test_pop_absent(self):
        # Non-existent name
        with self.assertRaises(KeyError):
            self.ss.pop('name4')
        # Non-existent with defaults
        self.assertIsNone(self.ss.pop('name4', None))
        self.assertEqual(self.ss.pop('name4', 'foo'), 'foo')
        # Remove one
        self.remove_callback.assert_not_called()
        self.assertIs(self.ss.pop('name0'), self.sensors[0])
        self._assert_sensors(self.ss, [])
        self.remove_callback.assert_called_once_with(self.sensors[0])

    def test_delitem(self):
        # Try to remove non-existent name
        with self.assertRaises(KeyError):
            del self.ss['name4']
        self._assert_sensors(self.ss, [self.sensors[0]])
        # Remove one
        self.remove_callback.assert_not_called()
        del self.ss['name0']
        self._assert_sensors(self.ss, [])
        self.remove_callback.assert_called_once_with(self.sensors[0])

    def test_getitem(self):
        # Non-existing name
        with self.assertRaises(KeyError):
            self.ss['name4']
        # Existing name
        self.assertIs(self.ss['name0'], self.sensors[0])

    def test_get(self):
        # Non-existing name
        self.assertIsNone(self.ss.get('name4'))
        self.assertIsNone(self.ss.get('name4', None))
        self.assertEqual(self.ss.get('name4', 'foo'), 'foo')
        # Existing name
        self.assertIs(self.ss.get('name0'), self.sensors[0])

    def test_len(self):
        self.assertEqual(len(self.ss), 1)
        self.ss.add(self.sensors[1])
        self.assertEqual(len(self.ss), 2)

    def test_contains(self):
        self.assertIn(self.sensors[0], self.ss)
        self.assertNotIn(self.alt_sensors[0], self.ss)
        self.assertNotIn(self.sensors[1], self.ss)

    def test_bool(self):
        self.assertTrue(self.ss)
        self.ss.clear()
        self.assertFalse(self.ss)

    def test_keys(self):
        self.ss.add(self.sensors[1])
        self.assertEqual(sorted(self.ss.keys()), ['name0', 'name1'])

    def test_values(self):
        self.ss.add(self.sensors[1])
        self.assertEqual(sorted(self.ss.values(), key=lambda x: x.name),
                         [self.sensors[0], self.sensors[1]])

    def test_items(self):
        self.ss.add(self.sensors[1])
        self.assertEqual(sorted(self.ss.items()), [
            ('name0', self.sensors[0]),
            ('name1', self.sensors[1])
        ])

    def test_iter(self):
        self.assertEqual(sorted(iter(self.ss)), ['name0'])

    def test_copy(self):
        self.assertEqual(self.ss.copy(), {'name0': self.sensors[0]})

    def test_remove_callbacks(self):
        self.ss.remove_remove_callback(self.remove_callback)
        self.ss.remove_add_callback(self.add_callback)
        self.ss.add(self.sensors[0])
        self.ss.remove(self.sensors[0])
        self.remove_callback.assert_not_called()
        self.add_callback.assert_not_called()



#!/usr/bin/env python3

# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import asyncio
import logging
import signal

import aiokatcp


async def main():
    client = await aiokatcp.Client.connect('localhost', 4444)
    async with client:
        while True:
            await client.wait_connected()
            value, _ = await client.request('echo', 10)
            print(value)
            await asyncio.sleep(1)


if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)
    loop = asyncio.get_event_loop()
    loop.set_debug(True)
    main_task = loop.create_task(main())
    loop.add_signal_handler(signal.SIGINT, main_task.cancel)
    try:
        loop.run_until_complete(main_task)
    except asyncio.CancelledError:
        pass
    loop.close()



# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import enum
import json
import ipaddress
from fractions import Fraction
import unittest
import unittest.mock
from typing import Union

from aiokatcp.core import (
    Message, KatcpSyntaxError, Address, Timestamp, TimestampOrNow, Now,
    encode, decode, register_type, get_type)


class MyEnum(enum.Enum):
    BATMAN = 1
    JOKER = 2
    TWO_FACE = 3


class MyIntEnum(enum.IntEnum):
    A = 1
    B = 2


class OverrideEnum(enum.Enum):
    BATMAN = b'cheese'
    JOKER = b'carrot'
    TWO_FACE = b'apple'

    def __init__(self, value):
        self.katcp_value = value


class TestAddress(unittest.TestCase):
    def setUp(self):
        self.v4_no_port = Address(ipaddress.ip_address('127.0.0.1'))
        self.v4_port = Address(ipaddress.ip_address('127.0.0.1'), 7148)
        self.v6_no_port = Address(ipaddress.ip_address('::1'))
        self.v6_port = Address(ipaddress.ip_address('::1'), 7148)
        self.v6_port_alt = Address(ipaddress.ip_address('00:00::1'), 7148)

    def test_getters(self) -> None:
        self.assertEqual(self.v4_no_port.host, ipaddress.ip_address('127.0.0.1'))
        self.assertIsNone(self.v4_no_port.port)
        self.assertEqual(self.v4_port.port, 7148)

    def test_eq(self) -> None:
        for addr in [self.v4_no_port, self.v4_port, self.v6_no_port, self.v6_port]:
            self.assertTrue(addr == addr)
        self.assertTrue(self.v6_port == self.v6_port_alt)
        self.assertFalse(self.v4_no_port == self.v4_port)
        self.assertFalse(self.v4_port == self.v6_port)
        self.assertFalse(self.v4_no_port == '127.0.0.1')

    def test_not_eq(self) -> None:
        self.assertFalse(self.v6_port != self.v6_port_alt)
        self.assertTrue(self.v4_no_port != self.v4_port)
        self.assertTrue(self.v4_no_port != self.v6_no_port)
        self.assertTrue(self.v4_no_port != '127.0.0.1')

    def test_str(self) -> None:
        self.assertEqual(str(self.v4_no_port), '127.0.0.1')
        self.assertEqual(str(self.v4_port), '127.0.0.1:7148')
        self.assertEqual(str(self.v6_no_port), '[::1]')
        self.assertEqual(str(self.v6_port), '[::1]:7148')

    def test_bytes(self) -> None:
        self.assertEqual(bytes(self.v4_no_port), b'127.0.0.1')
        self.assertEqual(bytes(self.v4_port), b'127.0.0.1:7148')
        self.assertEqual(bytes(self.v6_no_port), b'[::1]')
        self.assertEqual(bytes(self.v6_port), b'[::1]:7148')

    def test_repr(self) -> None:
        self.assertEqual(repr(self.v4_no_port), "Address(IPv4Address('127.0.0.1'))")
        self.assertEqual(repr(self.v4_port), "Address(IPv4Address('127.0.0.1'), 7148)")
        self.assertEqual(repr(self.v6_no_port), "Address(IPv6Address('::1'))")
        self.assertEqual(repr(self.v6_port), "Address(IPv6Address('::1'), 7148)")

    def test_hash(self) -> None:
        self.assertEqual(hash(self.v6_port), hash(self.v6_port_alt))
        self.assertNotEqual(hash(self.v4_port), hash(self.v4_no_port))
        self.assertNotEqual(hash(self.v4_port), hash(self.v6_port))

    def test_parse(self) -> None:
        for addr in [self.v4_no_port, self.v4_port, self.v6_no_port, self.v6_port]:
            self.assertEqual(Address.parse(bytes(addr)), addr)
        self.assertRaises(ValueError, Address.parse, b'')
        self.assertRaises(ValueError, Address.parse, b'[127.0.0.1]')
        self.assertRaises(ValueError, Address.parse, b'::1')


class TestEncodeDecode(unittest.TestCase):
    VALUES = [
        (str, 'café', b'caf\xc3\xa9'),
        (bytes, b'caf\xc3\xa9', b'caf\xc3\xa9'),
        (int, 123, b'123'),
        (bool, True, b'1'),
        (bool, False, b'0'),
        (float, -123.5, b'-123.5'),
        (float, 1e+20, b'1e+20'),
        (Fraction, Fraction(5, 4), b'1.25'),
        (Timestamp, Timestamp(123.5), b'123.5'),
        (TimestampOrNow, Timestamp(123.5), b'123.5'),
        (TimestampOrNow, Now.NOW, b'now'),
        (Address, Address(ipaddress.ip_address('127.0.0.1')), b'127.0.0.1'),
        (MyEnum, MyEnum.TWO_FACE, b'two-face'),
        (MyIntEnum, MyIntEnum.A, b'a'),
        (OverrideEnum, OverrideEnum.JOKER, b'carrot')
    ]

    BAD_VALUES = [
        (str, b'caf\xc3e'),      # Bad unicode
        (bool, b'2'),
        (int, b'123.0'),
        (float, b''),
        (Fraction, b'5/4'),
        (Address, b'[127.0.0.1]'),
        (MyEnum, b'two_face'),
        (MyEnum, b'TWO-FACE'),
        (MyEnum, b'bad-value'),
        (MyIntEnum, b'z'),
        (OverrideEnum, b'joker'),
        (TimestampOrNow, b'later'),
        (Union[int, float], b'123')
    ]

    def test_encode(self) -> None:
        for _, value, raw in self.VALUES:
            self.assertEqual(encode(value), raw)

    def test_decode(self) -> None:
        for type_, value, raw in self.VALUES:
            self.assertEqual(decode(type_, raw), value)

    def test_unknown_class(self) -> None:
        self.assertRaises(TypeError, decode, dict, b'{"test": "it"}')

    def test_bad_raw(self) -> None:
        for type_, value in self.BAD_VALUES:
            with self.assertRaises(ValueError,
                                   msg='{} should not be valid for {}'.format(value, type_)):
                decode(type_, value)

    @unittest.mock.patch('aiokatcp.core._types', [])   # type: ignore
    def test_register_type(self) -> None:
        register_type(
            dict, 'string',
            lambda value: json.dumps(value, sort_keys=True).encode('utf-8'),
            lambda cls, value: cls(json.loads(value.decode('utf-8'))))
        value = {"h": 1, "i": [2]}
        raw = b'{"h": 1, "i": [2]}'
        self.assertEqual(encode(value), raw)
        self.assertEqual(decode(dict, raw), value)
        self.assertRaises(ValueError, decode, dict, b'"string"')
        self.assertRaises(ValueError, decode, dict, b'{missing_quotes: 1}')
        # Try re-registering for an already registered type
        with self.assertRaises(ValueError):
            register_type(
                dict, 'string',
                lambda value: json.dumps(value, sort_keys=True).encode('utf-8'),
                lambda cls, value: cls(json.loads(value.decode('utf-8'))))

    def test_default(self) -> None:
        expected = [
            (int, 0),
            (float, 0.0),
            (str, ''),
            (bytes, b''),
            (Address, Address(ipaddress.IPv4Address('0.0.0.0'))),
            (Timestamp, Timestamp(0.0)),
            (MyEnum, MyEnum.BATMAN),
            (OverrideEnum, OverrideEnum.BATMAN)
        ]
        for type_, default in expected:
            self.assertEqual(get_type(type_).default(type_), default)


class TestMessage(unittest.TestCase):
    def test_init_basic(self) -> None:
        msg = Message(Message.Type.REQUEST,
                      'hello', 'world', b'binary\xff\x00', 123, 234.5, True, False)
        self.assertEqual(msg.mtype, Message.Type.REQUEST)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [
            b'world', b'binary\xff\x00', b'123', b'234.5', b'1', b'0'])
        self.assertIsNone(msg.mid)

    def test_init_mid(self) -> None:
        msg = Message(Message.Type.REPLY, 'hello', 'world', mid=345)
        self.assertEqual(msg.mtype, Message.Type.REPLY)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertEqual(msg.mid, 345)

    def test_init_bad_name(self) -> None:
        self.assertRaises(ValueError, Message, Message.Type.REPLY,
                          'underscores_bad', 'world', mid=345)
        self.assertRaises(ValueError, Message, Message.Type.REPLY,
                          '', 'world', mid=345)
        self.assertRaises(ValueError, Message, Message.Type.REPLY,
                          '1numberfirst', 'world', mid=345)

    def test_init_bad_mid(self) -> None:
        self.assertRaises(ValueError, Message, Message.Type.REPLY,
                          'hello', 'world', mid=0)
        self.assertRaises(ValueError, Message, Message.Type.REPLY,
                          'hello', 'world', mid=0x1000000000)

    def test_request(self) -> None:
        msg = Message.request('hello', 'world')
        self.assertEqual(msg.mtype, Message.Type.REQUEST)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertIsNone(msg.mid)

    def test_reply(self) -> None:
        msg = Message.reply('hello', 'world', mid=None)
        self.assertEqual(msg.mtype, Message.Type.REPLY)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertIsNone(msg.mid)

    def test_inform(self) -> None:
        msg = Message.inform('hello', 'world', mid=1)
        self.assertEqual(msg.mtype, Message.Type.INFORM)
        self.assertEqual(msg.name, 'hello')
        self.assertEqual(msg.arguments, [b'world'])
        self.assertEqual(msg.mid, 1)

    def test_reply_to_request(self) -> None:
        req = Message.request('hello', 'world', mid=1)
        reply = Message.reply_to_request(req, 'test')
        self.assertEqual(reply.mtype, Message.Type.REPLY)
        self.assertEqual(reply.name, 'hello')
        self.assertEqual(reply.arguments, [b'test'])
        self.assertEqual(reply.mid, 1)

    def test_inform_reply(self) -> None:
        req = Message.request('hello', 'world', mid=1)
        reply = Message.inform_reply(req, 'test')
        self.assertEqual(reply.mtype, Message.Type.INFORM)
        self.assertEqual(reply.name, 'hello')
        self.assertEqual(reply.arguments, [b'test'])
        self.assertEqual(reply.mid, 1)

    def test_bytes(self) -> None:
        msg = Message.request(
            'hello', 'café', b'_bin ary\xff\x00\n\r\t\\\x1b', 123, 234.5, True, False, '')
        raw = bytes(msg)
        expected = b'?hello caf\xc3\xa9 _bin\\_ary\xff\\0\\n\\r\\t\\\\\\e 123 234.5 1 0 \\@\n'
        self.assertEqual(raw, expected)

    def test_bytes_mid(self) -> None:
        msg = Message.reply('fail', 'on fire', mid=234)
        self.assertEqual(bytes(msg), b'!fail[234] on\\_fire\n')

    def test_repr(self) -> None:
        msg = Message.reply('fail', 'on fire', mid=234)
        self.assertEqual(repr(msg), "Message(Message.Type.REPLY, 'fail', b'on fire', mid=234)")

    def test_parse(self) -> None:
        msg = Message.parse(b'?test message \\0\\n\\r\\t\\e\\_binary\n')
        self.assertEqual(msg, Message.request('test', b'message', b'\0\n\r\t\x1b binary'))

    def test_parse_cr(self) -> None:
        msg = Message.parse(b'?test message withcarriagereturn\r')
        self.assertEqual(msg, Message.request('test', b'message', b'withcarriagereturn'))

    def test_parse_trailing_whitespace(self) -> None:
        msg = Message.parse(b'?test message  \n')
        self.assertEqual(msg, Message.request('test', b'message'))

    def test_parse_mid(self) -> None:
        msg = Message.parse(b'?test[222] message \\0\\n\\r\\t\\e\\_binary\n')
        self.assertEqual(msg, Message.request('test', b'message', b'\0\n\r\t\x1b binary', mid=222))
        msg = Message.parse(b'?test[1] message\n')
        self.assertEqual(msg, Message.request('test', b'message', mid=1))

    def test_parse_empty(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'')

    def test_parse_leading_whitespace(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b' !ok\n')

    def test_parse_bad_name(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'?bad_name message\n')
        self.assertRaises(KatcpSyntaxError, Message.parse, b'? emptyname\n')

    def test_parse_out_of_range_mid(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok[1000000000000]\n')

    def test_parse_bad_mid(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok[10\n')
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok[0]\n')
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok[a]\n')

    def test_parse_bad_type(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'%ok\n')

    def test_parse_bad_escape(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok \\q\n')
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok q\\ other\n')

    def test_parse_unescaped(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok \x1b\n')

    def test_parse_no_newline(self) -> None:
        self.assertRaises(KatcpSyntaxError, Message.parse, b'!ok')

    def test_compare(self) -> None:
        a = Message.request('info', 'yes')
        a2 = Message.request('info', 'yes')
        b = Message.request('info', 'yes', mid=123)
        c = Message.request('info', 'no')
        d = Message.request('other', 'yes')
        e = Message.reply('info', 'yes')
        f = 5
        for other in [b, c, d, e, f]:
            self.assertTrue(a != other)
            self.assertFalse(a == other)
        self.assertTrue(a == a2)

    def test_reply_ok(self) -> None:
        self.assertTrue(Message.reply('query', 'ok').reply_ok())
        self.assertFalse(Message.reply('query', 'fail', 'error').reply_ok())
        self.assertFalse(Message.request('query', 'ok').reply_ok())



#!/usr/bin/env python3

# Copyright 2019 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import asyncio
import logging
import signal

import aiokatcp


class MirrorServer(aiokatcp.DeviceServer):
    VERSION = 'mirror-1.0'
    BUILD_STATE = 'mirror-1.0.0'


class MirrorWatcher(aiokatcp.SensorWatcher):
    def __init__(self, client: aiokatcp.Client, server: aiokatcp.DeviceServer):
        super().__init__(client)
        self.server = server
        self._interface_stale = False

    def sensor_added(self, name: str, description: str, units: str, type_name: str,
                     *args: bytes) -> None:
        super().sensor_added(name, description, units, type_name, *args)
        self.server.sensors.add(self.sensors[name])
        self._interface_stale = True
        logging.info('Added sensor %s', name)

    def sensor_removed(self, name: str) -> None:
        del self.server.sensors[name]
        self._interface_stale = True
        super().sensor_removed(name)
        logging.info('Removed sensor %s', name)

    def batch_stop(self) -> None:
        if self._interface_stale:
            self.server.mass_inform('interface-changed', 'sensor-list')
            self._interface_stale = False
            logging.info('Sent interface-changed')


async def main(args: argparse.Namespace):
    client = aiokatcp.Client(args.host, args.port)
    server = MirrorServer('127.0.0.1', args.local_port)
    await server.start()

    watcher = MirrorWatcher(client, server)
    client.add_sensor_watcher(watcher)

    asyncio.get_event_loop().add_signal_handler(signal.SIGINT, server.halt)
    await server.join()
    client.close()
    await client.wait_closed()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description="Mirror another server's sensors")
    parser.add_argument('host')
    parser.add_argument('port', type=int)
    parser.add_argument('local_port', type=int)
    args = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main(args))
    loop.close()



#!/usr/bin/env python3

# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from setuptools import setup, find_packages

tests_require = ['asynctest', 'nose']
docs_require = ['sphinx', 'sphinx-autodoc-typehints', 'sphinxcontrib-asyncio', 'sphinx-rtd-theme']

setup(
    name='aiokatcp',
    use_katversion=True,
    packages=find_packages(),
    package_data={'aiokatcp': ['py.typed']},
    author='Bruce Merry',
    author_email='bmerry@ska.ac.za',
    description='Asynchronous I/O implementation of the katcp protocol',
    url='https://github.com/ska-sa/aiokatcp',
    license='BSD',
    platforms=['OS Independent'],
    keywords='asyncio katcp',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Framework :: AsyncIO',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Astronomy'
    ],
    tests_require=tests_require,
    install_requires=['decorator>=4.1', 'async-timeout'],
    setup_requires=['katversion'],
    extras_require={
        'test': tests_require,
        'doc': docs_require
    },
    entry_points={
        'console_scripts': [
            'katcpcmd = aiokatcp.tools.katcpcmd:main'
        ]
    },
    python_requires='>=3.5',
    zip_safe=False   # For py.typed
)



#!/usr/bin/env python3

# Copyright 2017 National Research Foundation (Square Kilometre Array)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
# list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Command-line request tool"""

import asyncio
import logging
import signal
import argparse
import contextlib

import async_timeout

import aiokatcp


def text(msg: aiokatcp.Message) -> str:
    return bytes(msg).decode('utf-8', errors='backslashreplace')


class CmdClient(aiokatcp.Client):
    def unhandled_inform(self, msg: aiokatcp.Message) -> None:
        print(text(msg), end='')


async def async_main(args, host, port) -> int:
    try:
        with async_timeout.timeout(args.connect_timeout):
            client = await CmdClient.connect(host, port, auto_reconnect=False)
    except (OSError, aiokatcp.client.ProtocolError) as error:
        logging.error('Connection error: %s', error)
        return 1
    except asyncio.TimeoutError:
        logging.error('Timed out connecting to %s:%s', host, port)
        return 1
    async with client:
        try:
            with async_timeout.timeout(args.request_timeout):
                reply, informs = await client.request_raw(args.command, *args.args)
        except (OSError, aiokatcp.client.ProtocolError) as error:
            logging.error('Connection error: %s', error)
            return 1
        except asyncio.TimeoutError:
            logging.error('Request timed out after %gs', args.request_timeout)
        else:
            for msg in informs + [reply]:
                print(text(msg), end='')
            if not reply.reply_ok():
                return 2
    return 0


def main() -> int:
    logging.basicConfig(level=logging.WARNING)

    parser = argparse.ArgumentParser(
        description='Send a single katcp request and print the reply',
        usage='%(prog)s host:port command [args...]')
    parser.add_argument('endpoint')
    parser.add_argument('command')
    parser.add_argument('args', nargs='*')
    parser.add_argument('--connect-timeout', type=float, metavar='TIME', default=30,
                        help='Time to wait for a connection to be established')
    parser.add_argument('--request-timeout', type=float, metavar='TIME',
                        help='Time to wait for the request to complete')
    args = parser.parse_args()
    host_port = args.endpoint.rsplit(':', 1)
    if len(host_port) != 2:
        parser.error('missing port number in {}'.format(host_port))

    loop = asyncio.get_event_loop()
    with contextlib.closing(loop):
        main_task = loop.create_task(async_main(args, *host_port))
        loop.add_signal_handler(signal.SIGINT, main_task.cancel)
        try:
            return loop.run_until_complete(main_task)
        except asyncio.CancelledError:
            return 1
